#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
  Code:        abq_part_elsets.py

  Author:      Jerawan Armstrong
               Monte Carlo Codes (XCP-3) Group
               X Computational Physics (XCP) Division
               Los Alamos National Laboratory

  Reference: This Python3 code was extracted from LA-UR-20-27139,
             approved for public release.

  Copyright (c) 2020 Triad National Security, LLC.  All rights reserved.

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE. 
"""
import numpy as np
from abaqus_input import extract_elset_data_generate
from abaqus_input import extract_elset_data

#------------------------------------------------------------------------------
def compute_part_elset_mat_nums(part_names, part_data, material_data):

    material_nums = material_data.keys()
    
    part_elset_mat_nums = {}
    
    for pname in part_names:
        elset_data = part_data.get(pname)[2]

        mat_nums = []
        
        for elset in elset_data:
            if 'material' not in elset[0]: continue

            elset_num = elset[1]
            
            if elset_num in material_nums:
                
                if elset_num not in mat_nums: mat_nums.append(elset_num)
                    
            else: 
                raise ValueError('elset material num {:d} in part {:s} \
is not in a list of materials'.format(elset[1],pname))

        if len(mat_nums) > 0:
            part_elset_mat_nums[pname] = mat_nums
        else:
            raise ValueError('no material for part {:s}'.format(pname))

    return part_elset_mat_nums

#------------------------------------------------------------------------------
def compute_part_elset_data(part_names, part_data):
    part_elset_data = {}
    
    for pname in part_names:
        pdata          = part_data.get(pname)
        elset_data     = pdata[2]
        mat_elsets     = pdata[3][0]
        matprop_elsets = pdata[3][1]
        stat_elsets    = pdata[3][2]

        if matprop_elsets >= stat_elsets:
            check_key    = 'matprop'
            n_check_keys = matprop_elsets
        else:
            check_key    = 'statistic'
            n_check_keys = stat_elsets

        if mat_elsets == 1 and n_check_keys == 1:
            complex_part_elset = False
        else:
            complex_part_elset = True

        order_elset_data     = []
        order_elset_numbers  = []

        for eldata in elset_data:
            add_data     = False
            elset_number = eldata[1]
                       
            for i in range(len(order_elset_numbers)):
                if elset_number <= order_elset_numbers[i]:
                    order_elset_data.insert(i,eldata)
                    order_elset_numbers.insert(i,elset_number)
                    add_data = True
                    break

            if not add_data:
                order_elset_data.append(eldata)
                order_elset_numbers.append(elset_number)

        if len(order_elset_data) != len(elset_data):
            raise ValueError('bad elset number for part {:>s}'.format(pname))

        else:
            part_elset_data[pname] = [order_elset_data, elset_data, check_key, mat_elsets, complex_part_elset]
            
    return part_elset_data


#------------------------------------------------------------------------------
def compute_part_elset_material_data(part_names, part_lines, part_elset_data, material_data):
    part_elset_material_data = {}
    names = []
    
    for pname in part_names:
        pelsetdata          = part_elset_data.get(pname)
        elset_data          = pelsetdata[1]
        check_key           = pelsetdata[2]
        complex_part_elset  = pelsetdata[4]

        # ---------------------------------------------------------------------
        # check for materials in a part that has complex elset lines
        if complex_part_elset:
            elset_lines  = part_lines.get(pname)[2]

            elset_mat_data = []
            for elset, lines in zip(elset_data,elset_lines):
                # elset[0] = list of elset keywords
                # elset[1] = elset num
                # elset[2] = genetate flag, =0 for w/o generate;  >0 for generate
                if 'material' in elset[0]:
                    if elset[2] >= 1:
                        line       = lines[0]
                        line_data  = extract_elset_data_generate(line)
                        this_elset_mat_data = set(np.arange(line_data[0],line_data[1]+1,line_data[2]))
                        
                    else:
                        this_elset_mat_data = []
                        for line in lines:
                            line_data  = extract_elset_data(line)
                            for n in line_data: this_elset_mat_data.append(n) 
                        this_elset_mat_data = set(this_elset_mat_data)
                
                    if len(this_elset_mat_data) == 0:
                        raise ValueError('no material elset data for elset number {:d} \
in part {:s}'.format(elset[1],pname))
                    else:
                        elset_mat_data.append([elset[1],this_elset_mat_data])
        # ------------------- end checking material ---------------------------

        for elset in elset_data:
            key_words = elset[0]
            if check_key not in key_words: continue
            
            elset_num = elset[1]
            
            name = '{:s}-{:d}'.format(pname,elset_num)
            if name not in names:
                names.append(name)
            else:                  
                raise ValueError('part {:s} has multiple {:s}-{:d} \
elset'.format(pname,check_key,elset_num))

            found_mat = False
            
            if not complex_part_elset:
                for t_elset in elset_data:
                    if 'material' not in t_elset[0]: continue
                    mat_num     = t_elset[1]
                    mat_data    = material_data.get(mat_num)
                    mat_name    = mat_data[0]
                    mat_density = mat_data[1]
                    found_mat   = True
                    break
                    
                if not found_mat:
                    raise ValueError('elements in {:s}-{:d} elset of part {:s} \
is not assigned to a material'.format(check_key,elset_num,pname))
                    
            
            else:
                elset_element_data = []
                
                for t_elset, lines in zip(elset_data,elset_lines):
                    # t_elset[0] = list of elset keywords
                    # t_elset[1] = elset num
                    # T_elset[2] = genetate flag, =0 for w/o generate;  >0 for generate
                    if check_key in t_elset[0] and elset_num == t_elset[1]:
                        if t_elset[2] >= 1:
                            line       = lines[0]
                            line_data  = extract_elset_data_generate(line)
                            elset_element_data = set(np.arange(line_data[0],line_data[1]+1,line_data[2]))
                        else:
                            for line in lines:
                                line_data  = extract_elset_data(line)
                                for n in line_data: elset_element_data.append(n) 
                            elset_element_data = set(elset_element_data)
                        break
                   
                if len(elset_element_data) == 0:
                    raise ValueError('no elset element data for {:s}-{:d} elset in part \
{:s}'.format(check_key,elset_num,pname))

                else:

                    for md  in elset_mat_data:
                        this_mat_num        = md[0]
                        this_elset_mat_data = md[1]
                        
                        if this_elset_mat_data.intersection(elset_element_data) == elset_element_data:
                            mat_num     = this_mat_num
                            mat_data    = material_data.get(mat_num)    
                            mat_name    = mat_data[0]
                            mat_density = mat_data[1]
                            found_mat   = True
                            break

                    if not found_mat:
                        raise ValueError('elements in {:s}-{:d} elset of part {:s} \
has an incorrect material assignment'.format(check_key,elset_num,pname))
                                   
            part_elset_material_data[name] = [mat_name, mat_num, mat_density]

    return part_elset_material_data

#-----------------------------------------------------------------------------------
def compute_part_elset_element_data(part_names, part_lines, part_elset_data,
                                    elset_type='unknown', n_source_keywords={}):
    
    part_elset_element_data = {}

    if elset_type=='source' and len(n_source_keywords) != len(part_names):
        raise ValueError('mismatch length for source elset')
    
    for pname in part_names:
        if elset_type=='source':
            if n_source_keywords.get(pname) == 0:
                name = pname+'0'
                part_elset_element_data[name] = [0, [], pname]
                continue
            
        pelsetdata        = part_elset_data.get(pname)
        elset_data        = pelsetdata[1]
        
        if elset_type == 'source':
            check_key = 'source'
        else:
            check_key = pelsetdata[2]
            
        elset_lines = part_lines.get(pname)[2]
        
        for elset in elset_data:
            key_words = elset[0]
            if check_key not in key_words: continue

            elset_num   = elset[1]
            name        = pname+'-'+str(elset_num)

            elset_element_data = []
            
            for this_elset, this_line in zip(elset_data, elset_lines):
                if check_key in this_elset[0] and elset_num == this_elset[1]:
                    if this_elset[2] >= 1:
                        line       = this_line[0]
                        line_data  = extract_elset_data_generate(line)
                        elset_element_data = list(np.arange(line_data[0],line_data[1]+1,line_data[2]))
                    else:
                        for line in this_line:
                            line_data  = extract_elset_data(line)
                            for n in line_data: elset_element_data.append(n)
                    break

            n_elements = len(elset_element_data)
            part_elset_element_data[name] = [n_elements, elset_element_data, pname]
        
    return part_elset_element_data

#-----------------------------------------------------------------------------------
